@@ -10,9 +10,10 @@ use strict;
 use Exporter;
 use vars qw(@ISA @EXPORT %dh);
 @ISA=qw(Exporter);
-@EXPORT=qw(&init &doit &complex_doit &verbose_print &error &warning &tmpdir
-	    &pkgfile &pkgext &pkgfilename &isnative &autoscript &filearray
-	    &filedoublearray &getpackages &basename &dirname &xargs %dh
+@EXPORT=qw(&init &doit &doit_noerror &complex_doit &verbose_print &error
+            &warning &tmpdir &pkgfile &pkgext &pkgfilename &isnative
+	    &autoscript &filearray &filedoublearray
+	    &getpackages &basename &dirname &xargs %dh
 	    &compat &addsubstvar &delsubstvar &excludefile &package_arch
 	    &is_udeb &udeb_filename &debhelper_script_subst &escape_shell
 	    &inhibit_log &load_log &write_log &commit_override_log
@@ -22,6 +23,9 @@ use vars qw(@ISA @EXPORT %dh);
 
 my $max_compat=10;
 
+# The Makefile changes this if debhelper is installed in a PREFIX.
+my $prefix="/usr";
+
 sub init {
 	my %params=@_;
 
@@ -91,7 +95,7 @@ sub init {
 	# the command line may affect it.
 	$dh{FIRSTPACKAGE}=${$dh{DOPACKAGES}}[0];
 
-	# If no error handling function was specified, just propigate
+	# If no error handling function was specified, just propagate
 	# errors out.
 	if (! exists $dh{ERROR_HANDLER} || ! defined $dh{ERROR_HANDLER}) {
 		$dh{ERROR_HANDLER}='exit \$?';
@@ -203,16 +207,25 @@ sub escape_shell {
 }
 
 # Run a command, and display the command to stdout if verbose mode is on.
+# Throws error if command exits nonzero.
+#
 # All commands that modifiy files in $TMP should be ran via this 
 # function.
 #
 # Note that this cannot handle complex commands, especially anything
 # involving redirection. Use complex_doit instead.
 sub doit {
+	doit_noerror(@_) || _error_exitcode(join(" ", @_));
+}
+
+sub doit_noerror {
 	verbose_print(escape_shell(@_));
 
 	if (! $dh{NO_ACT}) {
-		system(@_) == 0 || _error_exitcode(join(" ", @_));
+		return (system(@_) == 0)
+	}
+	else {
+		return 1;
 	}
 }
 
@@ -504,7 +517,8 @@ sub pkgfilename {
 # 1: package
 # 2: script to add to
 # 3: filename of snippet
-# 4: sed to run on the snippet. Ie, s/#PACKAGE#/$PACKAGE/
+# 4: either text: shell-quoted sed to run on the snippet. Ie, 's/#PACKAGE#/$PACKAGE/'
+#    or a sub to run on each line of the snippet. Ie sub { s/#PACKAGE#/$PACKAGE/ }
 sub autoscript {
 	my $package=shift;
 	my $script=shift;
@@ -521,11 +535,11 @@ sub autoscript {
 		$infile="$ENV{DH_AUTOSCRIPTDIR}/$filename";
 	}
 	else {
-		if (-e "/usr/share/debhelper/autoscripts/$filename") {
-			$infile="/usr/share/debhelper/autoscripts/$filename";
+		if (-e "$prefix/share/debhelper/autoscripts/$filename") {
+			$infile="$prefix/share/debhelper/autoscripts/$filename";
 		}
 		else {
-			error("/usr/share/debhelper/autoscripts/$filename does not exist");
+			error("$prefix/share/debhelper/autoscripts/$filename does not exist");
 		}
 	}
 
@@ -533,18 +547,34 @@ sub autoscript {
 	   && !compat(5)) {
 		# Add fragments to top so they run in reverse order when removing.
 		complex_doit("echo \"# Automatically added by ".basename($0)."\"> $outfile.new");
-		complex_doit("sed '$sed' $infile >> $outfile.new");
+		autoscript_sed($sed, $infile, "$outfile.new");
 		complex_doit("echo '# End automatically added section' >> $outfile.new");
 		complex_doit("cat $outfile >> $outfile.new");
 		complex_doit("mv $outfile.new $outfile");
 	}
 	else {
 		complex_doit("echo \"# Automatically added by ".basename($0)."\">> $outfile");
-		complex_doit("sed '$sed' $infile >> $outfile");
+		autoscript_sed($sed, $infile, $outfile);
 		complex_doit("echo '# End automatically added section' >> $outfile");
 	}
 }
 
+sub autoscript_sed {
+	my $sed = shift;
+	my $infile = shift;
+	my $outfile = shift;
+	if (ref($sed) eq 'CODE') {
+		open(IN, $infile) or die "$infile: $!";
+		open(OUT, ">>$outfile") or die "$outfile: $!";
+		while (<IN>) { $sed->(); print OUT }
+		close(OUT) or die "$outfile: $!";
+		close(IN) or die "$infile: $!";
+	}
+	else {
+		complex_doit("sed \"$sed\" $infile >> $outfile");
+	}
+}
+
 # Removes a whole substvar line.
 sub delsubstvar {
 	my $package=shift;
@@ -618,7 +648,7 @@ sub filedoublearray {
 	my $file=shift;
 	my $globdir=shift;
 
-	# executable confi files are a v9 thing.
+	# executable config files are a v9 thing.
 	my $x=! compat(8) && -x $file;
 	if ($x) {
 		require Cwd;
@@ -734,6 +764,44 @@ sub buildos {
 	}
 }
 
+# Passed a list of profiles to match against, returns true if
+# DEB_BUILD_PROFILES environment variable matched
+sub buildprofilesmatch {
+	my %debbuildprofiles = ();
+	if (exists $ENV{'DEB_BUILD_PROFILES'}) {
+		foreach my $profile (split(/\s+/, $ENV{'DEB_BUILD_PROFILES'})) {
+			$debbuildprofiles{$profile} = 1;
+		}
+	}
+
+	my $packageprofilesstr = shift;
+	my $package = shift;
+	my @packageprofiles = split(/\s+/, $packageprofilesstr);
+	my $err = sub { error("Build-Profiles field for package $package contains both positive and negative entries"); };
+	if ($#packageprofiles < 0 || $packageprofiles[0] =~ /^!/) {
+		# package profiles list is negative or empty
+		foreach my $packageprofile (@packageprofiles) {
+			$packageprofile =~ /^!(.*)$/ || &{$err}();
+			if ($debbuildprofiles{$1}) {
+				return 0;
+			}
+		}
+		return 1;
+	}
+	else {
+		# package profiles list is positive
+		foreach my $packageprofile (@packageprofiles) {
+			if ($packageprofile =~ /^!/) {
+				&{$err}();
+			}
+			if ($debbuildprofiles{$packageprofile}) {
+				return 1;
+			}
+		}
+		return 0;
+	}
+}
+
 # Returns source package name
 sub sourcepackage {
 	open (CONTROL, 'debian/control') ||
@@ -758,18 +826,20 @@ sub sourcepackage {
 # packages.
 # As a side effect, populates %package_arches and %package_types with the
 # types of all packages (not only those returned).
-my (%package_types, %package_arches);
+my (%package_types, %package_arches, %package_profiles);
 sub getpackages {
 	my $type=shift;
 
 	%package_types=();
 	%package_arches=();
+	%package_profiles=();
 	
 	$type="" if ! defined $type;
 
 	my $package="";
 	my $arch="";
 	my $package_type;
+	my $build_profiles;
 	my @list=();
 	my %seen;
 	open (CONTROL, 'debian/control') ||
@@ -787,6 +857,7 @@ sub getpackages {
 				error("debian/control has a duplicate entry for $package");
 			}
 			$package_type="deb";
+			$build_profiles="";
 		}
 		if (/^Architecture:\s*(.*)/) {
 			$arch=$1;
@@ -794,11 +865,15 @@ sub getpackages {
 		if (/^(?:X[BC]*-)?Package-Type:\s*(.*)/) {
 			$package_type=$1;
 		}
-		
+		if (/^Build-Profiles:\s*(.*)/) {
+		        $build_profiles=$1;
+		}
+
 		if (!$_ or eof) { # end of stanza.
 			if ($package) {
 				$package_types{$package}=$package_type;
 				$package_arches{$package}=$arch;
+				$package_profiles{$package}=$build_profiles;
 			}
 
 			if ($package &&
@@ -806,7 +881,8 @@ sub getpackages {
 			     (($type eq 'arch'  || $type eq 'both') && ($arch eq 'any' ||
 					     ($arch ne 'all' &&
 			                      samearch(buildarch(), $arch)))) ||
-			     ! $type)) {
+			     ! $type) &&
+			    buildprofilesmatch($build_profiles, $package)) {
 				push @list, $package;
 				$package="";
 				$arch="";
